/* Copyright Coraid, Inc. 2006.  All Rights Reserved. */
/* ethernet console for Coraid storage products */
/*  simple command line version */
/* <Tab 8> */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/errno.h>
#include <arpa/inet.h>
#include <ctype.h>
#include "cec.h"

#define	nelem(x)	(sizeof(x)/sizeof((x)[0]))
#define nil ((void *)0)


typedef struct Shelf Shelf;
typedef struct Pkt Pkt;

struct Pkt {
	uchar	dst[6];
	uchar	src[6];
	ushort	etype;
	uchar	type;
	uchar	conn;
	uchar	seq;
	uchar	len;
	uchar	data[1500];
};

enum {
	Tinita = 0,
	Tinitb,
	Tinitc,
	Tdata,
	Tack,
	Tdiscover,
	Toffer,
	Treset,
	
	HDRSIZ = 18,
};

struct Shelf {
	u_char	ea[6];
	int	shelfno;
	char	*str;
};

void 	probe(void);
int 	pickone(void);
void 	conn(int);
void 	gettingkilled(int);
void	sethdr(Pkt *, int);
int	shelfidx(void);

Shelf	tab[1000];
int	ntab;
u_char	contag;
extern 	int fd;	/* set in netopen */
int	shelf = -1;
Shelf	*connp;

void
usage(void)
{
	fprintf(stderr, "usage: %s [-d] [-s shelf] interface\n", progname);
	exit(1);
}

int
main(int argc, char **argv)
{
	int ch, r, n;
	
	progname = *argv;
	while ((ch = getopt(argc, argv, "ds:")) != -1) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 's':
			shelf = atoi(optarg);
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (debug)
		fprintf(stderr, "debug is on\n");
	if (argc != 1)
		usage();
	r = netopen(*argv);
	if (r == -1)
		fprintf(stderr, "%s: can't netopen %s\n", progname, *argv);
	probe();
loop:
	n = shelf == -1 ? pickone() : 0;
	rawon();
	signal(SIGTERM, gettingkilled);
	signal(SIGHUP, gettingkilled);
	signal(SIGKILL, gettingkilled);
	conn(n);
	rawoff();
	if (shelf != -1)
		return 0;
	shelf = -1;
	goto loop;
}

void
catch(int sig)
{
}

void
timewait(int secs)	/* arrange for a sig_alarm signal after `secs' seconds */
{
	struct sigaction sa;
	
	memset(&sa, 0, sizeof sa);
	sa.sa_handler = catch;
	sa.sa_flags = SA_RESETHAND;
	sigaction(SIGALRM, &sa, NULL);
	alarm(secs);
}

void
probe(void)
{
	u_char buf[1500];
	Pkt q;
	char *sh, *other;
	extern int errno;
	Shelf *p;
	int n;
	
	ntab = 0;
	memset(buf, 0xff, 6);
	memset(q.dst, 0xff, 6);
	memset(q.src, 0, 6);
	q.etype = htons(CEC_ETYPE);
	q.type = Tdiscover;
	q.len = 0;
	q.conn = 0;
	q.seq = 0;
	netsend(&q, 60);
	fprintf(stderr, "Probing for shelves ... ");
	fflush(stderr);
	timewait(WAITSECS);
	for (;;) {
		n = netrecv();
		if (n < 0 && errno == EINTR) {
			alarm(0);
			break;
		}
		while ((n = netget(&q, sizeof q)) > 0) {
			if (n < 60)
				continue;
			if (ntab >= nelem(tab))
				continue;
			if (memcmp(q.dst, "\xff\xff\xff\xff\xff\xff", 6) == 0)
				continue;
			if (q.len == 0)
				continue;
			q.data[q.len] = 0;
			sh = strtok((char *)q.data, " \t");
			if (sh == NULL)
				continue;
			if (shelf != -1 && atoi(sh) != shelf)
				continue;
			other = strtok(NULL, "\x1");
			p = tab+ntab++;
			memcpy(p->ea, q.src, 6);
			p->shelfno = atoi(sh);
			p->str = other ? strdup(other) : "";
			if (shelf != -1) {
				fprintf(stderr, "shelf %d found.\n", shelf);
				return;
			}
		}
	}
	if (ntab == 0) {
		fprintf(stderr, "none found.\n");
		exit(1);
	}
	printf("done.\n");
}

void
showtable(void)
{
	int i, j;

	printf("\n#  | SHELF | EA\n");
	for (i = 0; i < ntab; i++) {
		printf("%-2d   %-5d   ", i, tab[i].shelfno);
		for (j = 0; j < 6; j++)
			printf("%02x", tab[i].ea[j] & 0xff);
		printf("\n");
	}
}

int
pickone(void)
{
	char buf[80];
	int n, i;

	showtable();
	for (;;) {
		printf("[#qp]: ");
		fflush(stdout);
		n = read(0, buf, sizeof buf);
		switch (n) {
		case 1:
			if (buf[0] == '\n')
				continue;
		case 2:
			if (buf[0] == 'p') {
				probe();
				break;
			}
			if (buf[0] == 'q')
		case 0:
		case -1:
				exit(0);
		}
		if (isdigit(buf[0])) {
			buf[n] = 0;
			i = atoi(buf);
			if (i >= 0 && i < ntab)
				break;
		}
		showtable();
	}
	return i;
}

void
sethdr(Pkt *pp, int type)
{
	memmove(pp->dst, connp->ea, 6);
	memset(pp->src, 0, 6);
	pp->etype = htons(CEC_ETYPE);
	pp->type = type;
	pp->len = 0;
	pp->conn = contag;
}
	
void
ethclose(void)
{
	Pkt msg;
	
	sethdr(&msg, Treset);
	netsend(&msg, 60);
}

int
ethopen(void)
{
	Pkt pk;
	int cnt, n;
	
	contag = (getpid() >> 8) ^ (getpid() & 0xff);
	sethdr(&pk, Tinita);
	netsend(&pk, 60);
	
	/*  wait for INITB */
	
	fflush(stdout);
	timewait(WAITSECS);
	cnt = 3;
	do {
		n = netrecv();
		if (n < 0 && errno == EINTR) {
			alarm(0);
			return 0;
		}
		while ((n = netget(&pk, 1000)) > 0) {
			if (n < 0 && errno == EINTR) {
				if (--cnt > 0) {
					timewait(WAITSECS);
					netsend(&pk, 60);
				} else {
					alarm(0);
					return 0;
				}
			}
		}
	} while (pk.type != Tinitb);
	
	alarm(0);	
	sethdr(&pk, Tinitc);
	netsend(&pk, 60);
	return 1;
}

void
gettingkilled(int x)
{
	if (connp != nil)
		ethclose();
	exit(1);
}

void
prmem(char *label, void *p, int len)	/* debugging print */
{
	u_char *cp;
	int i;
	
	cp = (u_char *)p;
	printf("%s", label);
	for (i = 0; i < len; i++)
		printf("%02x", *cp++);
	printf("; ");
}

int
readln(int fd, char *buf, int len)
{
	int n;
	char ch;

	n = 0;
	while(n < len) {
		if(read(0, &ch, 1) != 1)
			return n;
		buf[n++] = ch;
		if(ch == '\n')
			break;
	}
	return n;
}

char
escape(void)
{
	char c, buf[64];
	int n;
loop:
	printf(">>> ");
	fflush(stdout);
	rawoff();
	n = readln(0, buf, sizeof buf - 1);
	rawon();
	if (n <= 0)
		return '.';
	c = buf[0];
	switch (c) {
	case 'i':
		return '';
	case 'q':
	case '.':
		return c;
	}
	printf("	(q)uit, (i)nterrupt, (.)continue\r\n");
	goto loop;
}

void
doloop(void)
{
	fd_set rfds;
	char c;
	int n, unacked = 0, retries, state;
	u_char sndseq = 0, rcvseq = -1;
	Pkt sndpkt, rcvpkt;
	u_char ea[6];
	struct timeval *tvp, timout;
	char esc = '';

	memmove(ea, connp->ea, 6);
	state = 0;
	for (;;) {
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		if (unacked == 0)
			FD_SET(0, &rfds);
		if (unacked) {
			tvp = &timout;
			tvp->tv_sec = 1;
			tvp->tv_usec = 0;
		} else
			tvp = NULL;
		n = select(fd+1, &rfds, nil, nil, tvp);
		if (n < 0) {
			perror("select failed");
			exit(8);
		}
		if (n == 0) { 	/* timeout */
			if (--retries == 0) {
				printf("Connection timed out\r\n");
				return;
			}
			netsend(&sndpkt, HDRSIZ + n < 60 ? 60 : HDRSIZ + n);
			continue;
		}
		if (FD_ISSET(0, &rfds)) {
			sethdr(&sndpkt, Tdata);
			n = read(0, &c, 1);
			if (n < 0) {
				perror("read failed");
				rawon();
				exit(8);
			}
			if (c == esc) {
				switch(c = escape()) {
				case 'q':
					sndpkt.len = 0;
					sndpkt.type = Treset;
					netsend(&sndpkt, 60);
					return;
				default:
					printf("unknown escape\r\n");
				case '.':
					continue;
				case '':
					break;
				}
			}
			sndpkt.data[0] = c;
			sndpkt.len = n;
			sndpkt.seq = ++sndseq;
			unacked = 1;
			retries = 3;
			netsend(&sndpkt, HDRSIZ + n < 60 ? 60 : HDRSIZ + n);
		} else if (FD_ISSET(fd, &rfds)) {
			n = netrecv();
			if (n < 0) {
				perror("netread failed");
				exit(8);
			}
			while ((n = netget(&rcvpkt, sizeof rcvpkt)) > 0) {
				if (n < 60)
					continue;
				if (memcmp(rcvpkt.src, ea, 6) != 0)
					continue;
				if (ntohs(rcvpkt.etype) != CEC_ETYPE)
					continue;
				if (rcvpkt.conn != contag)
					continue;
				switch (rcvpkt.type) {
				case Tinita: 
				case Tinitb: 
				case Tinitc:
				case Tdiscover:
				case Toffer:
					break;
				case Tdata:
					if (rcvpkt.seq == rcvseq)	/* ignore */
						break;
					write(1, rcvpkt.data, rcvpkt.len);
					
					/* ack data packet */
					
					memmove(rcvpkt.dst, rcvpkt.src, 6);
					memset(rcvpkt.src, 0, 6);
					rcvpkt.type = Tack;
					rcvpkt.len = 0;
					rcvseq = rcvpkt.seq;
					netsend(&rcvpkt, 60);
					break;
				case Tack:
					if (rcvpkt.seq == sndseq)
						unacked = 0;
					break;
				case Treset:
					return;
				}
			}
		}
	}
}

void
conn(int n)
{
	connp = &tab[n];
	printf("connecting ... ");
	if (ethopen() == 0) {
		printf("connection failed.\r\n");
		return;
	}
	printf("done.\r\n");
	printf("Escape is Ctrl-\\\r\n");
	doloop();
	ethclose();
}
